from typing import List, Dict
import tree_sitter
import tree_sitter_python as tspython
from tree_sitter import Language, Parser
import sys
sys.setrecursionlimit(200)

PY_LANGUAGE = Language(tspython.language())
language_parser = Parser(PY_LANGUAGE)

def traverse_type(node, results, kind, ignore_kind, avoid_nested= False) -> None:    
    if kind is None:
        results.append(node)
    elif node.type in kind:
        results.append(node)
        # avoid nested node (e.g. function, class) by not stop traveling if node type is found
        if avoid_nested:
            return 
    if not node.children:
        return
    for n in node.children:
        if n.type in ignore_kind:
            continue
        traverse_type(n, results, kind, ignore_kind, avoid_nested= avoid_nested)

def get_node_by_kind(root: tree_sitter.Node, kind: List[str], ignore_kind: List[str]=[], avoid_nested: bool= False) -> List:
    """
    Get all nodes with specific type
    
    Args:
        root (tree_sitter.Node): Tree sitter root node
        kind (List[str]): (node's) type that want to get
        ignore_kind (List[str]): (node's) type that DON'T want to get
        avoid_nested (bool): whether to avoid travel nested node with same type
    
    Return:
        List[tree_sitter.Node]: List of all 
    """

    node_list = []
    traverse_type(root, node_list, kind=kind, ignore_kind= ignore_kind, avoid_nested= avoid_nested)
    return node_list

def get_root_node(content: str) -> tree_sitter.Node:
    """
    Get the root node of the parsed content
    
    Args:
        content: input text content
    """
    root = language_parser.parse(bytes(content, "utf8"))
    return root.root_node

def get_import_nodes(content: str) -> List[tree_sitter.Node]:
    """
    Obtain all import statments in a code file
    
    Args:
        content: input text content
    
    Return:
        List of tree-sitter nodes that are import statement
    """
    root_node = get_root_node(content)
    import_nodes = get_node_by_kind(root_node, kind = ["import_statement", "import_from_statement", "future_import_statement"])
    return import_nodes

def parse_import(import_node):
    
    import_module_nodes = []
    package = None

    package_start = False
    module_start = False
    for children in import_node.children:
        if children.type == "from":
            package_start= True
            module_start = False
        elif package_start:
            assert children.type in ["dotted_name", "__future__", "relative_import"]
            package = children.text.decode()
            package_start = False
        elif children.type == "import":
            module_start = True
        elif module_start and children.type in ["dotted_name", "aliased_import"]:
            import_module_nodes.append(children)

    import_details = []
    for import_module_node in import_module_nodes:
        module_name = None
        import_module = import_module_node.text.decode()
        if import_module_node.type == "aliased_import":
            import_module, module_name = [x.strip() for x in import_module.split(" as ")]

        if package is not None:    
            import_text = "from {} import {}".format(package, import_module)
        else:
            import_text = "import {}".format(import_module)

        if module_name is not None:
            import_text += " as {}".format(module_name)

        import_details.append({"package": package, 
                                "module": import_module, 
                                "name": module_name if module_name else import_module, 
                                "import_text": import_text})
    return import_details

#### Code file cleaning
def decorated_clean(content: str) -> str:
    """
    Clean line decoration mark '@' (e.g. @staticmethod)

    Args:
        content: content of code file
    """
    lines = content.splitlines()
    new_lines = []
    for line in lines:
        if not line.strip().startswith("@"):
            new_lines.append(line)
    return "\n".join(new_lines)

def map_line_to_id(content: str) -> Dict:
    line2id = {}
    tmp = []
    for lid, line in enumerate(content.splitlines()):
        if lid == 0:
            line2id[lid] = 0
        else:
            line2id[lid] = len(bytes("\n".join([x.decode("utf8") for x in tmp]), "utf8")) + 1 # plus 1 extra new lines
        tmp.append(line)
    return line2id

def remove_content(content: str, rm_position: List) -> str:
    content = bytes(content, "utf8")
    line2id = map_line_to_id(content)
    new_rm_position = [0]
    for pos in rm_position:
        new_rm_position.extend([line2id[pos[0][0]]+ pos[0][1], line2id[pos[1][0]]+ pos[1][1]]) 
    new_rm_position.append(len(content))

    new_content= ''
    for pos_id in range(0, len(new_rm_position), 2):
        new_content += content[new_rm_position[pos_id]:new_rm_position[pos_id+1]].decode("utf8")
    return new_content

def remove_comment(content: str) -> str:
    root_node = get_root_node(content)

    positions = []
    for node in get_node_by_kind(root_node, kind= ["comment"]):
        positions.append((node.start_point, node.end_point))

    if positions:
        cleaned_content = remove_content(content, positions)
    else:
        cleaned_content = content
    return cleaned_content.strip()

def code_basic_clean(file_content: str) -> str:
    """
    1. clean highlevel comment
    2. clean decorated line (line startswith @ )
    3. clean main program - if __name__ == "__main__"
    """
    root_node = get_root_node(file_content)

    positions = []
    for node in root_node.children:
        if node.type == "comment" or (node.type == 'expression_statement' and node.children[0].type == 'string'):
            positions.append((node.start_point, node.end_point))

    if positions:
        cleaned_content = remove_content(bytes(file_content, "utf8"), positions)
    else:
        cleaned_content = file_content

    if "if __name__ == \"__main__\"" in cleaned_content:
        new_lines = cleaned_content.splitlines()
        for lid, line in enumerate(new_lines):
            if line.strip().startswith("if __name__ == \"__main__\""):
                break
        cleaned_content = "\n".join(new_lines[:lid])
    
    return decorated_clean(cleaned_content)



import os
from typing import List, Dict
from codetext.parser import PythonParser

from .build_repo_graph import get_repo_graph
from .utils import get_root_node, get_node_by_kind, get_import_nodes
from .Node import FunctionNode, ClassNode, BlockNode, ImportNode, ModuleNode
from .constant import PY_EXTENSIONS, EXCLUDED_TYPING_IDENTIFIERS
from .travel_graph import import_analyze

class Extractor:
    def __init__(self, repo_src: str, module: str=None):
        self.repo_src = repo_src
        self.module = module
        self.repo_graph = get_repo_graph(repo_src)

    def extract(self):
        if self.module is None:
            return self.repo_extract()
        else:
            return self.file_extract()

    def file_extract(self):
        functions = get_functions_from_module_file(self.module)
        import_nodes = get_import_from_module_file(self.module)
        module_function_dict = {"function": {}, "import": {}}
        module_node = ModuleNode(path= self.module)
        
        for function in functions:
            get_dependencies(function, module_function_dict, self.repo_src, self.repo_graph)
        
        for import_node in import_nodes:
            get_dependencies(import_node, module_function_dict, self.repo_src, self.repo_graph)

        if self.module in module_function_dict["function"]: # check no extracted functions
            module_node.function_list.extend(sorted([module_function_dict["function"][self.module][x] for x in module_function_dict["function"][self.module]], key= lambda x: x.position_in_file[0][0]))
        if self.module in module_function_dict["import"]: # check no extracted import
            module_node.import_list.extend(sorted([module_function_dict["import"][self.module][x] for x in module_function_dict["import"][self.module]], key= lambda x: x.position_in_file[0][0]))
        
        return module_node
        

    def repo_extract(self):
        all_modules = []
        get_modules_from_repo(self.repo_src, all_modules)
        module_function_dict = {"function": {}, "import": {}}
        module_dict = {}

        for module in all_modules:
            functions = get_functions_from_module_file(module)
            import_nodes = get_import_from_module_file(module)
            module_dict[module] = ModuleNode(path= module)

            for function in functions:
                get_dependencies(function, module_function_dict, self.repo_src, self.repo_graph)

            for import_node in import_nodes:
                get_dependencies(import_node, module_function_dict, self.repo_src, self.repo_graph)

            if module in module_function_dict["function"]:
                module_dict[module].function_list.extend(sorted([module_function_dict["function"][module][x] for x in module_function_dict["function"][module]], key= lambda x: x.position_in_file[0][0]))
            if module in module_function_dict["import"]:
                module_dict[module].import_list.extend(sorted([module_function_dict["import"][module][x] for x in module_function_dict["import"][module]], key= lambda x: x.position_in_file[0][0]))
        return module_dict

def get_modules_from_repo(repo_src: str, modules: List=[]) -> List[str]:
    """
    Get all module files from a repository

    Args:
        repo_src: local path of the repository
        modules: list of extracted modules
    """
    
    IGNORE_MODULES = (".git", "__pycache__")
    for sub_f in os.listdir(repo_src):
        if sub_f.startswith(IGNORE_MODULES):
            continue
        if os.path.isdir(os.path.join(repo_src, sub_f)):
            get_modules_from_repo(os.path.join(repo_src, sub_f), modules)
        
        for py_ext in PY_EXTENSIONS:
            if sub_f.endswith(py_ext) and not sub_f.startswith("__init__"):
                modules.append(os.path.join(repo_src, sub_f))
                break

def get_functions_from_module_file(module_path: str) -> List:
    extracted_functions = []

    # Read module file
    try:
        with open(module_path, "r", encoding="utf-8") as f:
            file_content = f.read()
    except:
        return []

    root_node = get_root_node(file_content)

    function_nodes = get_node_by_kind(root_node, kind= ["function_definition"], ignore_kind=["class_definition"], avoid_nested= True)
    if len(function_nodes) == 0:
        return []

    for function_node in function_nodes:
        extracted_functions.append(FunctionNode(module_path, function_node.text.decode(), function_node))

    return extracted_functions

def get_import_from_module_file(module_path: str) -> List:
    extracted_import = []

    # Read module file
    try:
        with open(module_path, "r", encoding="utf-8") as f:
            file_content = f.read()
    except:
        return []

    import_nodes = get_import_nodes(file_content)
    if len(import_nodes) == 0:
        return []

    for import_node in import_nodes:
        extracted_import.append(ImportNode(module_path, import_node.text.decode(), import_node))

    return extracted_import


def get_function_dependencies(target_function: FunctionNode, module_function_dict: Dict, repo_src: str, repo_graph: Dict) -> FunctionNode:
    
    local_path = target_function.path
    # Read module file
    try:
        with open(local_path, "r", encoding="utf-8") as f:
            file_content = f.read()
    except:
        return []

    root_node = get_root_node(file_content)

    # Obtain blocks that are function 
    for node in get_node_by_kind(root_node, kind= ["function_definition"], ignore_kind=["class_definition"], avoid_nested= True):
        iden_name = PythonParser.get_function_metadata(node, file_content)["identifier"]
        if iden_name in target_function.called_identifiers:
            if iden_name not in module_function_dict["function"][local_path]:
                dep_function = FunctionNode(local_path, node.text.decode(), node)
                get_dependencies(dep_function, module_function_dict, repo_src, repo_graph)
            target_function.children.append(module_function_dict["function"][local_path][iden_name])
    
    # Obtain blocks that are class 
    for node in get_node_by_kind(root_node, kind= ["class_definition"], ignore_kind=["function_definition"], avoid_nested= True):
        iden_name = PythonParser.get_function_metadata(node, file_content)["identifier"]
        if iden_name in target_function.called_identifiers:
            """
            TODO: update dependencies for class object
            # if iden_name not in module_function_dict[local_path]:
            #     dep_class = ClassNode(local_path, node.text.decode(), node)
            #     module_function_dict[local_path][iden_name] = get_dependencies(dep_function, module_function_dict)
            """
            target_function.children.append(ClassNode(local_path, node.text.decode(), node))

    # Obtain blocks that are variables, import, ...
    # Do we need to consider decorated_definition @ ?
    for node in root_node.children:
        if node.type in ["function_definition", "class_definition", "decorated_definition"]: # already consider function and class definition
            continue
        if "import" in node.type:
            node_identifiers = [x.text.decode() for x in get_node_by_kind(node, kind=["dotted_name"])]
            if "*" in node_identifiers or len(set(node_identifiers).intersection(set(target_function.called_identifiers))) > 0:
                if local_path not in module_function_dict["import"]:
                    module_function_dict["import"][local_path] = {}
                
                if node.text.decode() not in module_function_dict["import"][local_path]:
                    dep_import = ImportNode(local_path, node.text.decode(), node)
                    get_dependencies(dep_import, module_function_dict, repo_src, repo_graph)
                target_function.children.append(module_function_dict["import"][local_path][node.text.decode()])
        # only consider the first identifier
        # only consider "expression_statement" ?
        elif node.type in ["expression_statement"]:
            node_identifiers = [x.text.decode() for x in get_node_by_kind(node, kind=["identifier"])]
            if node_identifiers and node_identifiers[0] in target_function.called_identifiers and node_identifiers[0] not in EXCLUDED_TYPING_IDENTIFIERS: # avoid achieving redundant context with typing identifiers
                target_function.children.append(BlockNode(local_path, node.text.decode(), node))

    return target_function    

def get_import_dependencies(target_import: ImportNode, module_function_dict: Dict, repo_src: str, repo_graph: Dict) -> ImportNode:
    target_import.import_dict = import_analyze([target_import.tree_sitter_node], target_import.path, repo_graph)
    import_dict = [x for x in target_import.import_dict if x["import_path"] and x["import_path"].startswith(repo_src)]
    
    # third party import or import that are failed to obtain path
    if not import_dict:
        return target_import
    
    # Get related files
    selected_file = {}
    for import_info in import_dict:
        if import_info["import_path"] not in selected_file:
            selected_file[import_info["import_path"]] = []
        if import_info["name"] == "*" or import_info["import_file_or_folder"]:
            selected_file[import_info["import_path"]].append("*")
        else:
            selected_file[import_info["import_path"]].append(import_info["module"])
    
    for fid, import_file in enumerate(selected_file):
        try:
            # Sometime the import_path extracted is a directory. We ignore this case now :(
            with open(import_file, "r", encoding="utf-8") as f:
                cross_file_content = f.read()
        except:
            continue

        root_node = get_root_node(cross_file_content)

        # Obtain blocks that are function 
        for node in get_node_by_kind(root_node, kind= ["function_definition"], ignore_kind=["class_definition"], avoid_nested= True):
            iden_name = PythonParser.get_function_metadata(node)["identifier"]
            if "*" in selected_file[import_file] or iden_name in selected_file[import_file]:
                if import_file not in module_function_dict["function"]:
                    module_function_dict["function"][import_file] = {}

                if iden_name not in module_function_dict["function"][import_file]:
                    dep_function = FunctionNode(import_file, node.text.decode(), node)
                    get_dependencies(dep_function, module_function_dict, repo_src, repo_graph)
                target_import.children.append(module_function_dict["function"][import_file][iden_name])
        
        # Obtain blocks that are class 
        for node in get_node_by_kind(root_node, kind= ["class_definition"], ignore_kind=["function_definition"], avoid_nested= True):
            iden_name = PythonParser.get_function_metadata(node)["identifier"]
            if "*" in selected_file[import_file] or iden_name in selected_file[import_file]:
                """
                TODO: update dependencies for class object
                # if iden_name not in module_function_dict[local_path]:
                #     dep_class = ClassNode(local_path, node.text.decode(), node)
                #     module_function_dict[local_path][iden_name] = get_dependencies(dep_function, module_function_dict)
                """
                target_import.children.append(ClassNode(import_file, node.text.decode(), node))
            
        # Obtain blocks that are variables, ...
        # Do we need to consider decorated_definition @ ?
        for node in root_node.children:
            if node.type in ["function_definition", "class_definition", "decorated_definition"]:# or "import" in node.type: # already consider function and class definition
                continue
            
            if "import" in node.type:
                node_identifiers = [x.text.decode() for x in get_node_by_kind(node, kind=["dotted_name"])]
                if "*" in selected_file[import_file] or len(set(node_identifiers).intersection(set(selected_file[import_file]))) > 0:
                    if import_file not in module_function_dict["import"]:
                        module_function_dict["import"][import_file] = {}
                    
                    if node.text.decode() not in module_function_dict["import"][import_file]:
                        dep_import = ImportNode(import_file, node.text.decode(), node)
                        get_dependencies(dep_import, module_function_dict, repo_src, repo_graph)
                    target_import.children.append(module_function_dict["import"][import_file][node.text.decode()])
            elif node.type in ["expression_statement"]:
                node_identifiers = [x.text.decode() for x in get_node_by_kind(node, kind= ["identifier"])]
                if node_identifiers and node_identifiers[0] in selected_file[import_file]:
                    target_import.children.append(BlockNode(import_file, node.text.decode(), node))
    return target_import


def get_dependencies(target_node: FunctionNode | ImportNode, module_function_dict: Dict, repo_src: str, repo_graph: Dict):

    target_node_type = None
    if type(target_node) is FunctionNode:
        target_node_type = "function"
    elif type(target_node) is ImportNode:
        target_node_type = "import"
    else:
        raise ValueError("target_node type {} hasn't supported yet!".format(type(target_node)))
    

    if target_node.path not in module_function_dict[target_node_type]:
        module_function_dict[target_node_type][target_node.path] = {}

    if target_node_type == "function":
        if target_node.name in module_function_dict[target_node_type][target_node.path]:
            return
    elif target_node_type == "import":
        if target_node.content in module_function_dict[target_node_type][target_node.path]:
            return 
    
    try:
        # Some functions are recursive functions which can call forever
        # Break when reach max recursion => the tree could be very large
        if target_node_type == "function":
            target_node = get_function_dependencies(target_node, module_function_dict, repo_src, repo_graph)
        elif target_node_type == "import":
            target_node = get_import_dependencies(target_node, module_function_dict, repo_src, repo_graph)
    except Exception as e:
        # print(e)
        pass

    if target_node_type == "import":
        module_function_dict[target_node_type][target_node.path][target_node.content] = target_node
    elif target_node_type == "function":
        module_function_dict[target_node_type][target_node.path][target_node.name] = target_node


import os
import git
from typing import List
import re

def clone_repo(src: str, save_dir: str="repos"):
    """
    Clone repository if it is not downloaded

    Args:
        src: repo directory or link
        save_dir: save directory if cloning
    Return:
        Local directory of repository
    """
    if not os.path.exists(src):
        repo_path = os.path.join(save_dir, src.split("/")[-1].split(".")[0])
        
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        
        # Check if repository is already cloned
        if not os.path.exists(repo_path):
            git.Git(save_dir).clone(src)
        return repo_path
    else:
        return src

def fix_white_space(string: str):
    """Remove redundant white space"""
    return " ".join(string.split())

def remove_empty_line(string: str):
    """Remove empty line"""
    return "\n".join([x for x in string.splitlines() if x.strip()])

def find_all_substring(substring, string):
    return sorted([m.start() for m in re.finditer(substring, string)])


import json
from codetext.parser import PythonParser
from tree_sitter import Language, Parser
import tree_sitter
import os
from typing import Dict

from .utils import get_node_by_kind, get_root_node, decorated_clean
from .utils import language_parser as parser
from .constant import PY_EXTENSIONS

def get_identifier_in_file(filepath: str) -> Dict:
    """
    Get identifiers of a file

    Args:
        filepath: the file path to get identifiers
    """
    try:
        with open(filepath, "r") as f:
            filecontent = f.read()
        
        # Clean decorated line so easier to detect node type by 1st level
        filecontent = decorated_clean(filecontent)
    except Exception as e:
        # print(e)
        return {"path": filepath, "childrens": []}
        
    root_node = get_root_node(filecontent)

    all_modules = set()
    for children in root_node.children:
    
        if "import" in children.type:
            # Only consider the actual imported module
            # 1. import abc => abc
            # 2. from abc import xyz => xyz
            # 3. import abc as xyz => xyz

            import_as_identifiers = get_node_by_kind(children, kind=["aliased_import"])
            if import_as_identifiers:
                # import ... as ... or from .. import ... as ...
                all_modules.update([x.text.decode().split(" as ")[-1].strip() for x in import_as_identifiers])
            
            """
            Some cases could be:
            E.g.
                - from a import b, c as d
                - import a, b as c, d as e
            """
            start = False
            for subchild in children.children:
                if subchild.type == "import":
                    start = True
                elif start and subchild.type == "dotted_name":
                    all_modules.add(subchild.text.decode())
                
        else:
            child_identifiers = [x.text.decode() for x in get_node_by_kind(children, kind= ["identifier"])]
            # Obtain only the first occurrent identifier
            if child_identifiers:
                all_modules.add(child_identifiers[0])

    return {"path": filepath, "childrens": list(all_modules)}

    
def get_children(folder: str) -> Dict:
    graph_child = {"path": folder, "childrens": {}}
    for children in os.listdir(folder):
        if os.path.isdir(os.path.join(folder, children)):
            graph_child["childrens"][children] = get_children(os.path.join(folder, children))
        elif children.endswith(PY_EXTENSIONS):
            graph_child["childrens"][children] = get_identifier_in_file(os.path.join(folder, children))
        
    return graph_child        


def get_repo_graph(repo_src: str, save_graph_to: str= None) -> Dict:
    """
    Construct module graph for a repository

    Args:
        repo_src: the local path of the repository
        save_graph_to: directory to save the created graph
    """
    graph = get_children(repo_src)
    if save_graph_to:

        if not os.path.exists(save_graph_to):
            os.makedirs(save_graph_to)

        reponame = repo_src.split("/")[-1]
        with open(os.path.join(save_graph_to, f"{reponame}.json"), "w") as f:
            json.dump(graph, f, indent=4)
    return graph


import os

from .utils import parse_import, get_root_node, get_node_by_kind
from .constant import PY_EXTENSIONS

def search_path(key: str, graph, tracks= []):
    """
    Search the path from root of repo to the 'key' module. 
    
    Args:
        key (str): key module (starting point of a path) to be searched
        graph (Dict): graph of the repository
        tracks (List): tracking list of all searched path
    """
    if key is None:
        tracks.append(graph)
    else:
        for child in graph["childrens"]:
            if child == key or child in [key + x for x in PY_EXTENSIONS]:
                if type(graph["childrens"]) is not list:
                    tracks.append(graph["childrens"][child])
            else:
                if type(graph["childrens"]) is not list:
                    search_path(key, graph["childrens"][child], tracks= tracks)

def verify_track(separated_path, module, track):
    if len(separated_path) == 0:
        if type(track["childrens"]) is list:
            module_childrens = set(track["childrens"]).intersection(set([module + py_ext for py_ext in PY_EXTENSIONS]))
        else:
            module_childrens = set(track["childrens"].keys()).intersection(set([module + py_ext for py_ext in PY_EXTENSIONS]))
        module_childrens = list(module_childrens)
        assert len(module_childrens) <= 1, module_childrens

        if module not in track["childrens"] and \
        len(module_childrens) == 0:
            if "__init__.py" in track["childrens"]: # from folder(.__init__) import module
                if module not in track["childrens"]["__init__.py"]["childrens"]:
                    return False, False
                else:
                    return track["childrens"]["__init__.py"]["path"], False
            else:
                return False, False
        elif module in track["childrens"]:
            if type(track["childrens"]) is list:
                return track["path"], False
            # import folder
            return track["childrens"][module]["path"], True
        elif len(module_childrens) > 0:
            # import file
            if module + ".py" in module_childrens:
                return track["childrens"][module + ".py"]["path"], True
            return track["childrens"][module_childrens[0]]["path"], True
    else:
        if type(track["childrens"]) is list:
            module_childrens = set(track["childrens"]).intersection(set([separated_path[0] + py_ext for py_ext in PY_EXTENSIONS]))
        else:
            module_childrens = set(track["childrens"].keys()).intersection(set([separated_path[0] + py_ext for py_ext in PY_EXTENSIONS]))
        module_childrens = list(module_childrens)
        assert len(module_childrens) <= 1, module_childrens

        if separated_path[0] not in track["childrens"] and len(module_childrens) == 0:
            return False, False
        else:
            if separated_path[0] in track["childrens"]:
                if type(track["childrens"]) is list:
                    return False, False
                return verify_track(separated_path[1:], module, track["childrens"][separated_path[0]])
            else:
                if type(track["childrens"]) is list:
                    return False, False
                return verify_track(separated_path[1:], module, track["childrens"][module_childrens[0]])

def search_by_repo_graph(import_detail, repo_graph, current_path):
    def search_chain(separated_path, module, repo_graph, current_path):
        # Find all track
        all_tracks = []

        if len(separated_path) == 0: # import module
            # start_point = os.path.dirname(current_path).split("/")[-1]
            start_point = None
            separated_path= [start_point]
        else:
            start_point =  separated_path[0]
        search_path(start_point, repo_graph, all_tracks)

        for track in all_tracks:
            # start_point is always covered, skip 0 position
            track_verified, import_file_or_folder = verify_track(separated_path[1:], module, track)
            if track_verified:
                # if os.path.isdir(track_verified):
                #     print(separated_path, module, track_verified)
                return track_verified, import_file_or_folder
        return None, None
          
    if import_detail["package"] is not None:
        # import could be from fol1.fol2.file1 import 
        separated_path = import_detail["package"].split(".")
        track = search_chain(separated_path, import_detail["module"], repo_graph, current_path)

        if track is not None:
            return track
    else:
        separated_path = import_detail["module"].split(".")
        track = search_chain(separated_path[:-1], import_detail["module"].split(".")[-1], repo_graph, current_path)
        if track is not None:
            return track
    return None, None


def import_analyze(import_nodes, filepath, repo_graph):
    import_details = []
    for import_node in import_nodes:
        import_details.extend(parse_import(import_node))

    RELATIVE_IMPORT_PREFIXS = ["...", "..", "."]
    for import_detail in import_details:
        import_path = None
        import_file_or_folder = False

        for import_prefix in RELATIVE_IMPORT_PREFIXS:
            if import_detail["package"] is not None and import_detail["package"].startswith(import_prefix):
                start_dir = filepath
                for _ in range(len(import_prefix)):
                    start_dir = os.path.dirname(start_dir)
                
                if import_detail["package"] != import_prefix:
                    for py_ext in PY_EXTENSIONS:
                        if os.path.exists(os.path.join(start_dir, import_detail["package"][len(import_prefix):].replace(".", "/") + py_ext)): # from file import module 
                            import_path = os.path.join(start_dir, import_detail["package"][len(import_prefix):].replace(".", "/") + py_ext)
                            import_file_or_folder = False
                        elif os.path.exists(os.path.join(start_dir, import_detail["package"][len(import_prefix):].replace(".", "/"), import_detail["module"] + py_ext)): # from folder import file
                            import_path = os.path.join(start_dir, import_detail["package"][len(import_prefix):].replace(".", "/"), import_detail["module"] + py_ext)
                            import_file_or_folder = True
                        if import_path:
                            break
                    # Some functions might be imported from __init__ file
                    if not import_path:
                        init_path = os.path.join(start_dir, import_detail["package"][len(import_prefix):].replace(".", "/"), "__init__.py")
                        import_dir = os.path.join(start_dir, import_detail["package"][len(import_prefix):].replace(".", "/"), import_detail["module"])
                        if os.path.exists(init_path): # from folder import module in init file
                            with open(init_path, "r") as f:
                                init_content = f.read()
                            file_root_node= get_root_node(init_content)

                            if import_detail["module"] in [x.text.decode() for x in get_node_by_kind(file_root_node, kind=["identifier"])]:
                                import_path = init_path
                                import_file_or_folder = False
                        elif os.path.isdir(import_dir): # from folder import folder
                            # if os.path.exists(os.path.join(import_dir, "__init__.py")):
                            #     import_path = os.path.join(import_dir, "__init__.py")
                            import_path = import_dir
                            import_file_or_folder = True
                else:
                    for py_ext in PY_EXTENSIONS:
                        if os.path.exists(os.path.join(start_dir, import_detail["module"] + py_ext)):
                            import_path = os.path.join(start_dir, import_detail["module"] + py_ext) # from folder import file
                            import_file_or_folder = True
                        if import_path:
                            break
                        
                    if not import_path:
                        if os.path.isdir(os.path.join(start_dir, import_detail["module"])): # from folder import folder 
                            import_path = os.path.join(start_dir, import_detail["module"])
                            import_file_or_folder = True
                        elif os.path.exists(os.path.join(start_dir, "__init__.py")): # from folder import module in init file
                            with open(os.path.join(start_dir, "__init__.py"), "r") as f:
                                init_content = f.read()
                            file_root_node= get_root_node(init_content)

                            if import_detail["module"] in [x.text.decode() for x in get_node_by_kind(file_root_node, kind=["identifier"])]:
                                import_path = os.path.join(start_dir, "__init__.py")
                                import_file_or_folder = False
            if import_path:
                break
        
        if not import_path:
            import_path, import_file_or_folder = search_by_repo_graph(import_detail, repo_graph, filepath)


        import_detail["import_path"] = import_path
        import_detail["import_file_or_folder"] = import_file_or_folder
    return import_details


from pydepcall import Extractor
from pydepcall.Node import FunctionNode, ImportNode

src = "/home/namlh31aic/Project/AI4Code/pydepcall/tests/simple_repo"
module = "/home/namlh31aic/Project/AI4Code/pydepcall/tests/simple_repo/folder1/file2.py"

extractor = Extractor(src, None)
modules = extractor.extract()
module = modules[module]
print(len(module.import_list))

id = 1
function = module.function_list[id]

print(function.content)
print("="*100)
for child in function.children:
    print(child.content)
    if type(child) in [FunctionNode, ImportNode] and child.children:
        for x in child.children:
            print(x.content)
    print("-"*100)


from codetext.parser import PythonParser
import tree_sitter
from typing import List, Dict

from .travel_graph import import_analyze
from .utils import get_root_node, get_node_by_kind, remove_content, find_all_substring


class ModuleNode:
    """
    Module node object

    Attributes:
        path (str): the local path of file containing the function
        function list (list): list of functions (FunctionNode) in the module file
        import list (list): list of import statement (ImportNode) in the module file
    """
    def __init__(self, path: str):
        self.path = path
        self.function_list = []
        self.import_list = []
        


class FunctionNode:
    """
    Function node object

    Attributes:
        path (str): the local path of file containing the function
        content (str): the function content in text
        name (str): function's name
        params (dict): function's parameters
        return_type (str): function's return type
        docstring (str): function's docstring
        called_identifiers (list): called dependencies inside function
        position_in_file (tuple): start, end position of function in module file
        tree_sitter_node (tree_sitter.Node): Node parsed from file using tree-sitter
        children (list): List of called dependencies
    """
    def __init__(self, path: str, content: str, tree_sitter_node: tree_sitter.Node):
        self.path = path
        self.content = content
        self.tree_sitter_node = tree_sitter_node

        self.position_in_file = (tree_sitter_node.start_point, tree_sitter_node.end_point)
        self.name, self.params, self.return_type, self.docstring = self.get_metadata()
        self.called_identifiers = self.get_called_identifiers()

        self.children = []

    def get_metadata(self):
        function_metadata = PythonParser.get_function_metadata(self.tree_sitter_node)
        docstring = PythonParser.get_docstring(self.tree_sitter_node)
        return function_metadata["identifier"], function_metadata["parameters"], function_metadata["return_type"], docstring

    def get_called_identifiers(self):
        # Do not consider parameter identifier and typing types identifier. Do not forget to exclude self identifier
        identifiers = [x for x in get_dependencies(self.tree_sitter_node) if x not in list(self.params.keys()) + [self.name]]
        return identifiers
    

class ClassNode:
    """
    Class node object

    Attributes:
        path (str): the local path of file containing the function
        content (str): the function content in text
        name (str): function's name
        position_in_file (tuple): start, end position of class in module file
        tree_sitter_node (tree_sitter.Node): Node parsed from file using tree-sitter
    """
    def __init__(self, path: str, content: str, tree_sitter_node: tree_sitter.Node):
        self.path = path
        self.content = content    
        self.tree_sitter_node = tree_sitter_node

        self.position_in_file = (tree_sitter_node.start_point, tree_sitter_node.end_point)
        self.name = self.get_name()


    def get_name(self):
        return PythonParser.get_class_metadata(self.tree_sitter_node)["identifier"]

class BlockNode:
    """
    Block node object

    Attributes:
        path (str): the local path of file containing the function
        content (str): the function content in text
        name(str): first identifier's name
        position_in_file (tuple): start, end position of block in module file
        tree_sitter_node (tree_sitter.Node): Node parsed from file using tree-sitter
    """
    def __init__(self, path: str, content: str, tree_sitter_node: tree_sitter.Node):
        self.path = path
        self.content = content
        self.tree_sitter_node = tree_sitter_node
        self.name = self.get_name()
        self.position_in_file = (tree_sitter_node.start_point, tree_sitter_node.end_point)

    def get_name(self):
        return get_node_by_kind(self.tree_sitter_node, kind=["identifier"])[0].text.decode()

class ImportNode:
    """
    Import node object

    Attributes:
        path (str): the local path of file containing the function
        content (str): the function content in text
        position_in_file (tuple): start, end position of import statement in module file
        tree_sitter_node (tree_sitter.Node): Node parsed from file using tree-sitter
        children (list): List of called dependencies
    """
    def __init__(self, path: str, content: str, tree_sitter_node: tree_sitter.Node):
        self.path = path
        self.content = content
        self.tree_sitter_node = tree_sitter_node

        self.position_in_file = (tree_sitter_node.start_point, tree_sitter_node.end_point)
        self.import_dict = []
        self.children = []



def remove_parentheses(input_string):
    
    root_node = get_root_node(input_string)
    rm_positions = []
    
    # Remove string so that we can avoid parentheses in string
    for node in get_node_by_kind(root_node, kind=["string"], avoid_nested=True):
        rm_positions.append((node.start_point, node.end_point)) 
    string = remove_content(input_string, rm_positions)

    # Remove all parentheses
    open_indexes = find_all_substring("\(", string)
    close_indexes = find_all_substring("\)", string)
    assert len(open_indexes) == len(close_indexes), input_string
    modified_string = string[0:open_indexes[0]]
    for cid, close_index in enumerate(close_indexes):
        if cid == len(close_indexes) - 1:
            modified_string += string[close_index+1:]
        else:
            modified_string += string[close_index+1:open_indexes[cid+1]]
    return modified_string

def get_dependencies(function_node):
    """
    
    """

    must_contain_identifiers = set()
    exclude_identifiers = set()

    for call_node in get_node_by_kind(function_node, kind=["call"]):
        call_content = remove_parentheses(call_node.text.decode())
        identifiers = call_content.split(".")

        if len(identifiers) == 1:
            must_contain_identifiers.add(identifiers[0])
        else:
            exclude_identifiers.update(identifiers[1:])
    
    included_identifiers = set()
    for identifier_node in get_node_by_kind(function_node, kind=["identifier"]):
        identifier_text = identifier_node.text.decode()
        if identifier_text in must_contain_identifiers or identifier_text not in exclude_identifiers:
            included_identifiers.add(identifier_text)
    return included_identifiers

    
    






from .utils import clone_repo, remove_empty_line, fix_white_space, find_all_substring
from .parser_utils import (language_parser, get_node_by_kind, get_root_node, \
    code_basic_clean, remove_comment, remove_content, get_import_nodes, parse_import, decorated_clean)


import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "pydepcall",
    version = "0.0.2",
    author = "Nam Le Hai",
    author_email = "nam.lh173264@gmail.com",
    description = "Extract dependencies of modules (function, class, import) in python repository",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/FSoft-AI4Code/pydepcall",
    project_urls = {
        "repository": "https://github.com/FSoft-AI4Code/pydepcall",
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    python_requires = ">=3.10"
)
